/*
 *  Copyright 2001-2005 Adrian Thurston <thurston@cs.queensu.ca>
 */

/*  This file is part of Kelbt.
 *
 *  Kelbt is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  Kelbt is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with Kelbt; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

%{

#include <iostream>
#include <stdlib.h>
#include <limits.h>
#include <errno.h>
#include "kelbt.h"
#include "parsedata.h"

using std::endl;

/* Bison 1.875 tries to prevent a warning, but results ends up producing a 
 * This is an ugly hack. But it works for this case. */
#ifdef BISON_V1875
#define __attribute__(arg)
#endif

InputData *id = 0;
int includeDepth = 0;

typedef AvlMap<char*, ParseData *, CmpStr> SectionMap;
typedef AvlMapEl<char*, ParseData *> SectionMapEl;
SectionMap sectionMap;
LangElVect langElVect;
FactorVect factorVect;

/* These come from the scanner and point back into the parser. We will borrow
 * them for error reporting. */
extern YYSTYPE *yylval;
extern YYLTYPE *yylloc;

/* The include stack pointer from the scanner. Used to determine if we are
 * currently processing an included file. */
extern int inc_stack_ptr;

/* Try to do a definition, common to assignment and instantiation. */
void beginOutsideCode();
bool getParseData( const BISON_YYLTYPE &loc );
void endSection( const InputLoc &loc );
void doInclude( const InputLoc &loc, char *sectionName, char *inputFile );
int yylex( YYSTYPE *yylval, YYLTYPE *yylloc );

%}

%pure-parser

%union {
	/* General data types. */
	char c;
	char *data;
	bool boolean;
	Literal *literal;

	/* Tree nodes. */
	Factor *factor;
	ProdElList *prodElList;

	/* Priorities and actions. */
	InlineBlock *inlineBlock;
	LangEl *langEl;

	/* Inline parse tree items. */
	InlineItem *inlineItem;
	InlineList *inlineList;
}

/* Outside Code. */
%token OC_Struct
%token OC_Open
%token OC_Close
%token OC_Semi
%token OC_Interface
%token OC_Implementation
%token OC_Protocol
%token OC_End
%token <data> OC_FsmName

%token TK_StartSection
%token TK_EndSection

/* General tokens. */
%token <data> TK_Word
%token <data> TK_Literal

/* Tokens specific to inline code. */
%token <data> IL_Text
%token <data> IL_Reference

/* Keywords. */
%token KW_Init
%token KW_Parser
%token KW_AlphType
%token KW_GetKey
%token KW_Context
%token KW_Interface
%token KW_Include
%token KW_Element
%token KW_Token
%token KW_NonTerm
%token KW_Translate
%token KW_Try
%token KW_Final
%token KW_PreReduce
%token KW_PostReduce
%token KW_Type
%token KW_Uses
%token KW_Undo
%token KW_Commit
%token KW_TrailingCommit

/* Classes of transitions and states. */
%token KW_Enter
%token KW_All
%token KW_Finish
%token KW_Leave
%token KW_Start
%token KW_Final

/* Types of embeddings. */
%token KW_Exec
%token KW_Pri
%token KW_On
%token KW_Ctx
%token KW_Into
%token KW_From
%token KW_EOF
%token KW_Err
%token KW_LErr

/* Special token for terminating semi-terminated code blocks. Needed because
 * semi is sent as a token in the code block rather than as a generic symbol. */
%token TK_Semi

/* Symbols. In kelbt lexical space, the scanner does not pass 
 * any data along with the symbols, in inline code lexical 
 * space it does. */
%token '*' '?' '+' '!' '^' '(' ')' ';' ',' '=' 
%token ':' '@' '%' '$' '-' '|' '&' '>'

%type <boolean> OptTrailingCommit
%type <boolean> OptCommit
%type <inlineBlock> OptReduction
%type <inlineBlock> OptTranslate
%type <prodElList> ProdElList
%type <factor> Factor
%type <data> MachineName
%type <data> SectionName
%type <data> OptSection
%type <data> OptFileName
%type <inlineList> InlineBlock
%type <inlineItem> InlineBlockItem
%type <inlineBlock> OptUndoBlock
%type <inlineBlock> OptFinalBlock


%%

/* Input is any number of input sections. An empty file is accepted. */
input: FsmSpecList;
FsmSpecList: 		
	FsmSpecList FsmSpec |
	/* Nothing */;

/* Fsm Specification. Fsms are begin with '%%' and may be a {} delimited
 * list of Fsm statements or may be a single statement. If no name is
 * given the last name given in a machine is used. */
FsmSpec: 
	StartSection SectionName StatementList TK_EndSection {
		if ( id->active )
			endSection( id->sectionEndLoc );
	};

StartSection: 
	TK_StartSection { 
		/* Clear the last section name. */
		id->active = true;
		id->sectionName = 0;
		id->lookedUpSection = false;
		id->sawInterface = false;
		id->sawStart = false;
		id->sectionLoc = InputLoc(@1);
	};

SectionName: 
	KW_Parser TK_Word ';' {
		/* Used by the parse data structure lookup. */
		id->sectionName = $2;
	};

/* A NonEmpty list of statements in a fsm. */
StatementList:
	StatementList Statement |
	/* Nothing */;

/* The differnt types of statements in a fsm spec. */
Statement:	
	Production |
	InitSpec |
	Interface |
	Include |
	TokenList |
	ElementBlock |
	TokenBlock |
	TypeBlock |
	NonTermBlock |
	TranslateBlock |
	NonTermUses |
	PreReduceBlock |
	PostReduceBlock;

/* Garble up to the next ; */
Statement: error ';' { yyerrok; };

TokenList: 
	KW_Token WordList ';' {
		if ( includeDepth == 0 )
			id->pd->bAnyTokens = true;
	};

WordList: 
	WordList ',' TokenName |
	TokenName;

TokenName: 
	TK_Word {
		if ( getParseData( @1 ) ) {
			/* Token name must be unique. */
			LangEl *tokName = getLangEl( id->pd, $1, LangEl::Term );
			if ( tokName->type != LangEl::Unknown && tokName->type != LangEl::Term )
				error(@1) << "'" << $1 << "' already defined as something else" << endl;
			tokName->type = LangEl::Term;
		}
	};

ElementBlock: 
	KW_Element '{' InlineBlock '}' ';' {
		if ( getParseData( @1 ) )
			id->pd->elementBlock = new InlineBlock( InputLoc(@2), $3 );
	};

TokenBlock: 
	KW_Token '{' InlineBlock '}' ';' {
		if ( getParseData( @1 ) )
			id->pd->tokenBlock = new InlineBlock( InputLoc(@2), $3 );
	};

TypeBlock: 
	KW_Type TK_Word '{' InlineBlock '}' ';' {
		if ( getParseData( @1 ) ) {
			TypeDef *newTypeDef = new TypeDef( $2, new InlineBlock( InputLoc(@3), $4 ) );
			id->pd->typeDefMap.insert( $2, newTypeDef );
		}
	};

NonTermBlock: 
	KW_NonTerm TK_Word '{' InlineBlock '}' ';' {
		if ( getParseData( @1 ) ) {
			LangEl *nonTerm = getLangEl( id->pd, $2, LangEl::NonTerm );
			nonTerm->typeDef = new TypeDef( $2, new InlineBlock( InputLoc(@3), $4 ) );
			id->pd->typeDefMap.insert( $2, nonTerm->typeDef );
		}
	};

TranslateBlock: 
	KW_Token OptTranslate OptUndoBlock OptFinalBlock ';' {
		if ( getParseData( @1 ) ) {
			id->pd->translateBlock = $2;
			id->pd->undoTransBlock = $3;
			id->pd->tokenFinalBlock = $4;
		}
	};

NonTermUses:
	KW_NonTerm TK_Word KW_Uses TK_Word ';' {
		if ( getParseData( @1 ) ) {
			LangEl *nonTerm = getLangEl( id->pd, $2, LangEl::NonTerm );
			TypeDefMapEl *tdmel = id->pd->typeDefMap.find( $4 );
			if ( tdmel != 0 )
				nonTerm->typeDef = tdmel->value;
			else {
				error(@1) << "bad uses statement" << endl;
			}
		}
	}

PreReduceBlock: 
	KW_PreReduce '{' InlineBlock '}' ';' {
		if ( getParseData( @1 ) )
			id->pd->preReduceBlock = new InlineBlock( InputLoc(@2), $3 );
	};

PostReduceBlock: 
	KW_PostReduce '{' InlineBlock '}' ';' {
		if ( getParseData( @1 ) )
			id->pd->postReduceBlock = new InlineBlock( InputLoc(@2), $3 );
	};


/* Allow the user to give some code to insert into the init function. */
InitSpec:
	KW_Init '{' InlineBlock '}' {
		if ( getParseData( @1 ) ) {
			/* Put the code on the init code list. */
			id->pd->initCodeList.append( new InlineBlock( InputLoc(@2), $3 ) );
		}
		id->sectionEndLoc = @4;
	};

Interface:
	KW_Interface ';' {
		if ( getParseData( @1 ) ) {
			id->sawInterface = true;
		}
		id->sectionEndLoc = @2;
	}

/* Include statements are processed by both the scanner and the parser. */
Include:
	IncludeKeyword OptSection OptFileName ';' {
		if ( id->active )
			doInclude( @1, $2, $3 );
	}

IncludeKeyword: 
	KW_Include {
		/* Do this immediately so that the scanner has a correct sense of the
		 * value in id->active when it reaches the end of the statement before
		 * the above action executes. */
		getParseData( @1 );
	};

OptSection: TK_Word { $$ = $1; } | { $$ = 0; };
OptFileName: TK_Literal { $$ = $1; } | { $$ = 0; };

ProdElList:
	ProdElList Factor {
		$$ = $1;
		$$->append( $2 );
	} |
	/* Empty */ {
		$$ = new ProdElList();
	};

Production:
	MachineName ':' ProdElList OptTrailingCommit OptReduction 
	OptUndoBlock OptFinalBlock ';' {
		if ( getParseData( @2 ) ) {
			/* Get the language element. */
			LangEl *prodName = getLangEl( id->pd, $1, LangEl::NonTerm );

			/* Check that the element wasn't previously defined as something else. */
			if ( prodName->type != LangEl::Unknown 
					&& prodName->type != LangEl::NonTerm )
			{
				error(@1) << "'" << $1 << "' already defined as something else" << endl;
			}
			else {
				prodName->type = LangEl::NonTerm;
				Definition *newDef = new Definition( @1, prodName, $3,
						$4, $5, $6, $7, id->pd->prodList.length(), Definition::Production );
				prodName->defList.append( newDef );
				id->pd->prodList.append( newDef );

				/* Start machine will trigger code generation. */
				if ( strcmp($1, startDefName) == 0 )
					id->sawStart = true;
				id->pd->bAnyDefinitions = true;

				newDef->rhsLelVect = langElVect;
				newDef->rhsFactorVect = factorVect;
				langElVect.empty();
				factorVect.empty();
			}
			id->sectionEndLoc = @8;
		}
	};

OptTrailingCommit:
	/* Empty */ { $$ = false; } |
	KW_TrailingCommit { $$ = true; };

OptTranslate:
	/* Empty */ {
		$$ = 0;
	} |
	KW_Translate '{' InlineBlock '}' {
		$$ = new InlineBlock( InputLoc(@2), $3 );
	};

OptReduction:
	/* Empty */ {
		$$ = 0;
	} |
	KW_Try '{' InlineBlock '}' {
		$$ = new InlineBlock( InputLoc(@2), $3 );
	};

OptUndoBlock:
	KW_Undo '{' InlineBlock '}' {
		$$ = new InlineBlock( InputLoc(@2), $3 );
	} |
	/* Nothing */ {
		$$ = 0;
	};

OptFinalBlock: 
	KW_Final '{' InlineBlock '}' {
		$$ = new InlineBlock( InputLoc(@2), $3 );
	} |
	/* Nothing */ {
		$$ = 0;
	};

/* Capture the machine name for making the machine's priority name. */
MachineName:
	TK_Word {
		if ( getParseData( @1 ) ) {
		}
	};


/* The highest level in precedence. Atomic machines such as references to other
 * machines, literal machines, regular expressions or Expressions inside of
 * parenthesis. */
Factor:
	OptCommit TK_Literal {
		// Create a new factor node going to a concat literal. */
		assert( $2[0] == 0 || $2[1] ==  0 );
		$$ = new Factor( $1, new Literal( InputLoc(@2), $2) );
		langElVect.append( 0 );
		factorVect.append( $$ );
	} |
	OptCommit TK_Word {
		if ( id->active ) {
			LangEl *langEl = getLangEl( id->pd, $2, LangEl::Unknown );
			$$ = new Factor( @2, $1, langEl );
			langElVect.append( langEl );
			factorVect.append( $$ );
		}
	};

OptCommit: 
	/* Nothing */ { $$ = false; } |
	KW_Commit { $$ = true; }; 

InlineBlock:
	InlineBlock InlineBlockItem {
		/* Append the item to the list, return the list. */
		$1->append( $2 );
		$$ = $1;
	} |
	/* Empty */ {
		/* Start with empty list. */
		$$ = new InlineList;
	};

InlineBlockItem:
	IL_Text {
		/* Add a text segment. */
		$$ = new InlineItem( @1, $1, InlineItem::Text );
	} |
	IL_Reference {
		/* Add a text segment. */
		$$ = new InlineItem( @1, $1, InlineItem::Reference );
	};

%%

void doInclude( const InputLoc &loc, char *sectionName, char *inputFile )
{
	/* Bail if we hit the max include depth. */
	if ( includeDepth == INCLUDE_STACK_SIZE ) {
		error(loc) << "hit maximum include depth of " << INCLUDE_STACK_SIZE << endl;
	}
	else {
		char *includeTo = id->pd->fsmName;

		/* Implement defaults for the input file and section name. */
		if ( inputFile == 0 ) 
			inputFile = id->fileName;
		if ( sectionName == 0 )
			sectionName = id->pd->fsmName;

		/* Parse the included file. */
		InputData *oldId = id;
		id = new InputData( inputFile, sectionName, includeTo );
		includeDepth += 1;
		yyparse();
		includeDepth -= 1;
		delete id;
		id = oldId;
	}
}

bool getParseData( const BISON_YYLTYPE &loc )
{
	/* Lookup only once. */
	if ( id->lookedUpSection )
		return id->active;

	/* If we don't have a section name then try to get one from the structure
	 * that we are in. */
	if ( id->sectionName == 0 ) {
		/* If we don't yet have a section name, try to get the fsm spec name from
		 * any structure we are contained in. */
		if ( id->structStack.length() > 0 ) {
			char *head = id->structStack[id->structStack.length()-1];
			if ( head != 0 )
				id->sectionName = head;
		}
	}

	/* If we have a structure name then use that to get a 
	 * parse data structure. */
	if ( id->sectionName != 0 ) {
		if ( id->includeSpec != 0 ) {
			if ( strcmp( id->sectionName, id->includeSpec ) == 0 )
				id->sectionName = id->includeTo;
			else
				id->active = false;
		}

		/* Lookup the parse data, if it is not there then create it. */
		SectionMapEl *sectionMapEl = sectionMap.find( id->sectionName );
		if ( sectionMapEl == 0 ) {
			ParseData *newPd = new ParseData( id->fileName, id->sectionName, 
					id->sectionLoc );
			sectionMapEl = sectionMap.insert( id->sectionName, newPd );
		}
		id->pd = sectionMapEl->value;
	}
	else { 
		/* We don't have a name to use, only option is to use the last parse
		 * data. If there was no last named parse data then this is an error. */
		if ( id->pd == 0 ) {
			error(loc) << "don't have a name for this FSM specification" << endl;
			id->active = false;
		}
	}
	id->lookedUpSection = true;
	return id->active;
}

/* Invoked at the end of a section. Decided what to output depending on the
 * mode that kelbt is invoked in (graphviz, code) and the state of the parse. */
void endSection( const InputLoc &loc )
{
	/* In order to generate anything we must be in the top level file and the
	 * current spec must be active and there must not have been any parse
	 * errors. */
	if ( inc_stack_ptr > 0 || !id->active || id->pd->errorCount > 0 )
		return;

	int endLine = id->sectionEndLoc.line;

	if ( id->sawInterface ) {
		if ( ! ::generateGraphviz )
			id->pd->generateHeader( endLine );
	}

	if ( id->sawStart ) {
		if ( ::generateGraphviz )
			id->pd->generateGraphviz();
		else 
			id->pd->generateCode( id->sawInterface, endLine );
	}
}


void yyerror( char *err )
{
	/* Bison won't give us the location, but in the last call to the scanner we
	 * saved a pointer to the locationn variable. Use that. instead. */
	error(::yylloc->first_line, ::yylloc->first_column) << err << endl;
}

